@@ -1,5 +1,5 @@
 /* Copyright (c) 2000, 2010, Oracle and/or its affiliates.
-   Copyright (c) 2011, 2013, Monty Program Ab.
+   Copyright (c) 2011, 2020, MariaDB Corporation.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -12,7 +12,7 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA */
+   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1335  USA */
 
 /* The hash functions used for saveing keys */
 /* One of key_length or key_length_offset must be given */
@@ -45,7 +45,7 @@ my_hash_value_type my_hash_sort(CHARSET_INFO *cs, const uchar *key,
                                 size_t length)
 {
   ulong nr1= 1, nr2= 4;
-  cs->coll->hash_sort(cs, (uchar*) key, length, &nr1, &nr2);
+  my_ci_hash_sort(cs, (uchar*) key, length, &nr1, &nr2);
   return (my_hash_value_type) nr1;
 }
 
@@ -60,8 +60,9 @@ my_hash_value_type my_hash_sort(CHARSET_INFO *cs, const uchar *key,
   dynamic array that is part of the hash will allocate memory
   as required during insertion.
 
+  @param[in]     psi_key      The key to register instrumented memory
   @param[in,out] hash         The hash that is initialized
-  @param[in[     growth_size  size incrememnt for the underlying dynarray
+  @param[in]     growth_size  size incrememnt for the underlying dynarray
   @param[in]     charset      The character set information
   @param[in]     size         The hash size
   @param[in]     key_offest   The key offset for the hash
@@ -76,14 +77,14 @@ my_hash_value_type my_hash_sort(CHARSET_INFO *cs, const uchar *key,
     @retval 1 failure
 */
 my_bool
-my_hash_init2(HASH *hash, uint growth_size, CHARSET_INFO *charset,
-              ulong size, size_t key_offset, size_t key_length,
-              my_hash_get_key get_key,
+my_hash_init2(PSI_memory_key psi_key, HASH *hash, uint growth_size,
+              CHARSET_INFO *charset, ulong size, size_t key_offset,
+              size_t key_length, my_hash_get_key get_key,
               my_hash_function hash_function,
               void (*free_element)(void*), uint flags)
 {
   my_bool res;
-  DBUG_ENTER("my_hash_init");
+  DBUG_ENTER("my_hash_init2");
   DBUG_PRINT("enter",("hash:%p  size: %u", hash, (uint) size));
 
   hash->records=0;
@@ -95,7 +96,7 @@ my_hash_init2(HASH *hash, uint growth_size, CHARSET_INFO *charset,
   hash->free=free_element;
   hash->flags=flags;
   hash->charset=charset;
-  res= init_dynamic_array2(&hash->array, sizeof(HASH_LINK), NULL, size,
+  res= init_dynamic_array2(psi_key, &hash->array, sizeof(HASH_LINK), NULL, size,
                            growth_size, MYF((flags & HASH_THREAD_SPECIFIC ?
                                              MY_THREAD_SPECIFIC : 0)));
   DBUG_RETURN(res);
@@ -116,17 +117,23 @@ my_hash_init2(HASH *hash, uint growth_size, CHARSET_INFO *charset,
 static inline void my_hash_free_elements(HASH *hash)
 {
   uint records= hash->records;
+  if (records == 0)
+    return;
+
   /*
     Set records to 0 early to guard against anyone looking at the structure
     during the free process
   */
   hash->records= 0;
+
   if (hash->free)
   {
     HASH_LINK *data=dynamic_element(&hash->array,0,HASH_LINK*);
     HASH_LINK *end= data + records;
-    while (data < end)
+    do
+    {
       (*hash->free)((data++)->data);
+    } while (data < end);
   }
 }
 
@@ -748,7 +755,7 @@ my_bool my_hash_update(HASH *hash, uchar *record, uchar *old_key,
 }
 
 
-uchar *my_hash_element(HASH *hash, ulong idx)
+uchar *my_hash_element(HASH *hash, size_t idx)
 {
   if (idx < hash->records)
     return dynamic_element(&hash->array,idx,HASH_LINK*)->data;
@@ -786,14 +793,13 @@ void my_hash_replace(HASH *hash, HASH_SEARCH_STATE *current_record,
 my_bool my_hash_iterate(HASH *hash, my_hash_walk_action action, void *argument)
 {
   uint records, i;
-  HASH_LINK *data;
 
   records= hash->records;
-  data= dynamic_element(&hash->array,0,HASH_LINK*);
 
   for (i= 0 ; i < records ; i++)
   {
-    if ((*action)(data[i].data, argument))
+    if ((*action)(dynamic_element(&hash->array, i, HASH_LINK *)->data,
+                  argument))
       return 1;
   }
   return 0;
@@ -872,8 +878,8 @@ my_bool my_hash_check(HASH *hash)
 
 #define RECORDS 1000
 
-uchar *test_get_key(uchar *data, size_t *length,
-                    my_bool not_used __attribute__((unused)))
+const uchar *test_get_key(const void *data, size_t *length,
+                          my_bool not_used __attribute__((unused)))
 {
   *length= 2;
   return data;
@@ -889,8 +895,8 @@ int main(int argc __attribute__((unused)),char **argv __attribute__((unused)))
   DBUG_PUSH("d:t:O,/tmp/test_hash.trace");
 
   printf("my_hash_init\n");
-  if (my_hash_init2(&hash_test, 100, &my_charset_bin, 20,
-                    0, 0, (my_hash_get_key) test_get_key, 0, 0, HASH_UNIQUE))
+  if (my_hash_init2(PSI_INSTRUMENT_ME, &hash_test, 100, &my_charset_bin, 20, 0,
+                    0, test_get_key, 0, 0, HASH_UNIQUE))
   {
     fprintf(stderr, "hash init failed\n");
     exit(1);
