@@ -12,7 +12,7 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA */
+   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1335  USA */
 
 /*
   More functions to be used with IO_CACHE files
@@ -23,51 +23,56 @@
 #include <stdarg.h>
 #include <m_ctype.h>
 
-/*
-  Copy contents of an IO_CACHE to a file.
-
-  SYNOPSIS
-    my_b_copy_to_file()
-    cache  IO_CACHE to copy from
-    file   File to copy to
-
-  DESCRIPTION
-    Copy the contents of the cache to the file. The cache will be
-    re-inited to a read cache and will read from the beginning of the
-    cache.
-
-    If a failure to write fully occurs, the cache is only copied
-    partially.
+/**
+  Copy the cache to the file. Copying can be constrained to @c count
+  number of bytes when the parameter is less than SIZE_T_MAX.  The
+  cache will be optionally re-inited to a read cache and will read
+  from the beginning of the cache.  If a failure to write fully
+  occurs, the cache is only copied partially.
 
   TODO
-    Make this function solid by handling partial reads from the cache
-    in a correct manner: it should be atomic.
+  Make this function solid by handling partial reads from the cache
+  in a correct manner: it should be atomic.
 
-  RETURN VALUE
-    0  All OK
-    1  An error occurred
+  @param cache      IO_CACHE to copy from
+  @param file       File to copy to
+  @param count      the copied size or the max of the type
+                    when the whole cache is to be copied.
+  @return
+         0          All OK
+         1          An error occurred
 */
 int
-my_b_copy_to_file(IO_CACHE *cache, FILE *file)
+my_b_copy_to_file(IO_CACHE *cache, FILE *file,
+                  size_t count)
 {
-  size_t bytes_in_cache;
+  size_t curr_write, bytes_in_cache;
   DBUG_ENTER("my_b_copy_to_file");
 
-  /* Reinit the cache to read from the beginning of the cache */
-  if (reinit_io_cache(cache, READ_CACHE, 0L, FALSE, FALSE))
-    DBUG_RETURN(1);
   bytes_in_cache= my_b_bytes_in_cache(cache);
   do
   {
-    if (my_fwrite(file, cache->read_pos, bytes_in_cache,
+    curr_write= MY_MIN(bytes_in_cache, count);
+    if (my_fwrite(file, cache->read_pos, curr_write,
                   MYF(MY_WME | MY_NABP)) == (size_t) -1)
       DBUG_RETURN(1);
-  } while ((bytes_in_cache= my_b_fill(cache)));
+
+    cache->read_pos += curr_write;
+    count -= curr_write;
+  } while (count && (bytes_in_cache= my_b_fill(cache)));
   if(cache->error == -1)
     DBUG_RETURN(1);
   DBUG_RETURN(0);
 }
 
+int my_b_copy_all_to_file(IO_CACHE *cache, FILE *file)
+{
+  DBUG_ENTER("my_b_copy_all_to_file");
+  /* Reinit the cache to read from the beginning of the cache */
+  if (reinit_io_cache(cache, READ_CACHE, 0L, FALSE, FALSE))
+    DBUG_RETURN(1);
+  DBUG_RETURN(my_b_copy_to_file(cache, file, SIZE_T_MAX));
+}
 
 my_off_t my_b_append_tell(IO_CACHE* info)
 {
