@@ -11,7 +11,7 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA */
+   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1335  USA */
 
 /* Not MT-SAFE */
 
@@ -63,6 +63,8 @@ void* my_once_alloc(size_t Size, myf MyFlags)
       return((uchar*) 0);
     }
     DBUG_PRINT("test",("my_once_malloc %lu byte malloced", (ulong) get_size));
+    my_once_allocated+= get_size;
+    update_malloc_size(get_size, 0);
     next->next= 0;
     next->size= get_size;
     next->left= get_size-ALIGN_SIZE(sizeof(USED_MEM));
@@ -113,7 +115,9 @@ void my_once_free(void)
     old=next; next= next->next ;
     free((uchar*) old);
   }
+  update_malloc_size(- (longlong) my_once_allocated, 0);
   my_once_root_block=0;
+  my_once_allocated= 0;
 
   DBUG_VOID_RETURN;
 } /* my_once_free */
