@@ -18,6 +18,7 @@
 #include "log.h"
 #include "sql_plugin.h"
 #include <my_crypt.h>
+#include <violite.h>
 
 /* there can be only one encryption plugin enabled */
 static plugin_ref encryption_manager= 0;
@@ -58,11 +59,14 @@ uint ctx_size(unsigned int, unsigned int)
 
 } /* extern "C" */
 
-int initialize_encryption_plugin(st_plugin_int *plugin)
+int initialize_encryption_plugin(void *plugin_)
 {
+  st_plugin_int *plugin= static_cast<st_plugin_int *>(plugin_);
   if (encryption_manager)
     return 1;
 
+  vio_check_ssl_init();
+
   if (plugin->plugin->init && plugin->plugin->init(plugin))
   {
     sql_print_error("Plugin '%s' init function returned error.",
@@ -104,8 +108,9 @@ int initialize_encryption_plugin(st_plugin_int *plugin)
   return 0;
 }
 
-int finalize_encryption_plugin(st_plugin_int *plugin)
+int finalize_encryption_plugin(void *plugin_)
 {
+  st_plugin_int *plugin= static_cast<st_plugin_int *>(plugin_);
   bool used= plugin_ref_to_int(encryption_manager) == plugin;
 
   if (used)
