@@ -2,6 +2,7 @@
 #define ITEM_XMLFUNC_INCLUDED
 
 /* Copyright (c) 2000, 2019, Oracle and/or its affiliates. All rights reserved.
+   Copyright (c) 2009, 2019, MariaDB
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -23,6 +24,42 @@
 typedef struct my_xml_node_st MY_XML_NODE;
 
 
+/* Structure to store nodeset elements */
+class MY_XPATH_FLT
+{
+public:
+  uint num;     // Absolute position in MY_XML_NODE array
+  uint pos;     // Relative position in context
+  uint size;    // Context size
+public:
+  MY_XPATH_FLT(uint32 num_arg, uint32 pos_arg)
+   :num(num_arg), pos(pos_arg), size(0)
+  { }
+  MY_XPATH_FLT(uint32 num_arg, uint32 pos_arg, uint32 size_arg)
+   :num(num_arg), pos(pos_arg), size(size_arg)
+  { }
+  bool append_to(Native *to) const
+  {
+    return to->append((const char*) this, (uint32) sizeof(*this));
+  }
+};
+
+
+class NativeNodesetBuffer: public NativeBuffer<16*sizeof(MY_XPATH_FLT)>
+{
+public:
+  const MY_XPATH_FLT &element(uint i) const
+  {
+    const MY_XPATH_FLT *p= (MY_XPATH_FLT*) (ptr() + i * sizeof(MY_XPATH_FLT));
+    return *p;
+  }
+  uint32 elements() const
+  {
+    return length() / sizeof(MY_XPATH_FLT);
+  }
+};
+
+
 class Item_xml_str_func: public Item_str_func
 {
 protected:
@@ -73,16 +110,16 @@ protected:
 public:
   Item_xml_str_func(THD *thd, Item *a, Item *b): Item_str_func(thd, a, b)
   {
-    maybe_null= TRUE;
+    set_maybe_null();
   }
   Item_xml_str_func(THD *thd, Item *a, Item *b, Item *c):
     Item_str_func(thd, a, b, c)
   {
-    maybe_null= TRUE;
+    set_maybe_null();
   }
-  bool fix_fields(THD *thd, Item **ref);
-  bool fix_length_and_dec();
-  bool const_item() const
+  bool fix_fields(THD *thd, Item **ref) override;
+  bool fix_length_and_dec() override;
+  bool const_item() const override
   {
     return const_item_cache && (!nodeset_func || nodeset_func->const_item());
   }
@@ -94,25 +131,34 @@ class Item_func_xml_extractvalue: public Item_xml_str_func
 public:
   Item_func_xml_extractvalue(THD *thd, Item *a, Item *b):
     Item_xml_str_func(thd, a, b) {}
-  const char *func_name() const { return "extractvalue"; }
-  String *val_str(String *);
-  Item *get_copy(THD *thd)
+  LEX_CSTRING func_name_cstring() const override
+  {
+    static LEX_CSTRING name= {STRING_WITH_LEN("extractvalue") };
+    return name;
+  }
+  String *val_str(String *) override;
+  Item *do_get_copy(THD *thd) const override
   { return get_item_copy<Item_func_xml_extractvalue>(thd, this); }
 };
 
 
 class Item_func_xml_update: public Item_xml_str_func
 {
-  String tmp_value2, tmp_value3;
+  NativeNodesetBuffer tmp_native_value2;
+  String tmp_value3;
   bool collect_result(String *str,
                       const MY_XML_NODE *cut,
                       const String *replace);
 public:
   Item_func_xml_update(THD *thd, Item *a, Item *b, Item *c):
     Item_xml_str_func(thd, a, b, c) {}
-  const char *func_name() const { return "updatexml"; }
-  String *val_str(String *);
-  Item *get_copy(THD *thd)
+  LEX_CSTRING func_name_cstring() const override
+  {
+    static LEX_CSTRING name= {STRING_WITH_LEN("updatexml") };
+    return name;
+  }
+  String *val_str(String *) override;
+  Item *do_get_copy(THD *thd) const override
   { return get_item_copy<Item_func_xml_update>(thd, this); }
 };
 
