@@ -24,6 +24,7 @@
 #include "my_json_writer.h"
 #include "opt_range.h"
 #include "sql_expression_cache.h"
+#include "item_subselect.h"
 
 const char * STR_DELETING_ALL_ROWS= "Deleting all rows";
 const char * STR_IMPOSSIBLE_WHERE= "Impossible WHERE";
@@ -161,7 +162,7 @@ void Explain_query::query_plan_ready()
   Send EXPLAIN output to the client.
 */
 
-int Explain_query::send_explain(THD *thd)
+int Explain_query::send_explain(THD *thd, bool extended)
 {
   select_result *result;
   LEX *lex= thd->lex;
@@ -174,8 +175,22 @@ int Explain_query::send_explain(THD *thd)
   if (thd->lex->explain_json)
     print_explain_json(result, thd->lex->analyze_stmt);
   else
+  {
     res= print_explain(result, lex->describe, thd->lex->analyze_stmt);
-
+    if (extended)
+    {
+      char buff[1024];
+      String str(buff,(uint32) sizeof(buff), system_charset_info);
+                 str.length(0);
+     /*
+       The warnings system requires input in utf8, @see
+        mysqld_show_warnings().
+     */
+     lex->unit.print(&str, QT_EXPLAIN_EXTENDED);
+                     push_warning(thd, Sql_condition::WARN_LEVEL_NOTE,
+                     ER_YES, str.c_ptr_safe());
+    }
+  }
   if (res)
     result->abort_result_set();
   else
@@ -185,6 +200,7 @@ int Explain_query::send_explain(THD *thd)
 }
 
 
+
 /*
   The main entry point to print EXPLAIN of the entire query
 */
@@ -192,6 +208,9 @@ int Explain_query::send_explain(THD *thd)
 int Explain_query::print_explain(select_result_sink *output, 
                                  uint8 explain_flags, bool is_analyze)
 {
+  /* A sanity check for ANALYZE: */
+  DBUG_ASSERT(timer_tracker_frequency() != 0.0);
+
   if (upd_del_plan)
   {
     upd_del_plan->print_explain(this, output, explain_flags, is_analyze);
@@ -296,7 +315,7 @@ static void push_string_list(THD *thd, List<Item> *item_list,
     else
       buf->append(',');
 
-    buf->append(line);
+    buf->append(line, strlen(line));
   }
   push_string(thd, item_list, buf);
 }
@@ -406,7 +425,7 @@ int print_explain_row(select_result_sink *result,
       Item_float *fl= new (mem_root) Item_float(thd, *r_rows, 2);
       String tmp;
       String *res= fl->val_str(&tmp);
-      r_rows_str.append(res->ptr());
+      r_rows_str.append(*res);
       item_list.push_back(new (mem_root)
                           Item_string_sys(thd, r_rows_str.ptr(),
                                           r_rows_str.length()), mem_root);
@@ -552,7 +571,7 @@ int Explain_union::print_explain(Explain_query *query,
     Item_float *fl= new (mem_root) Item_float(thd, avg_rows, 2);
     String tmp;
     String *res= fl->val_str(&tmp);
-    r_rows_str.append(res->ptr());
+    r_rows_str.append(*res);
     item_list.push_back(new (mem_root)
                         Item_string_sys(thd, r_rows_str.ptr(),
                                         r_rows_str.length()), mem_root);
@@ -662,7 +681,11 @@ int Explain_node::print_explain_for_children(Explain_query *query,
   for (int i= 0; i < (int) children.elements(); i++)
   {
     Explain_node *node= query->get_node(children.at(i));
-    if (node->print_explain(query, output, explain_flags, is_analyze))
+    /*
+      Note: node may not be present because for certain kinds of subqueries,
+      the optimizer is not able to see that they were eliminated.
+    */
+    if (node && node->print_explain(query, output, explain_flags, is_analyze))
       return 1;
   }
   return 0;
@@ -706,8 +729,15 @@ void Explain_node::print_explain_json_for_children(Explain_query *query,
   for (int i= 0; i < (int) children.elements(); i++)
   {
     Explain_node *node= query->get_node(children.at(i));
-    /* Derived tables are printed inside Explain_table_access objects */
     
+    /*
+      Note: node may not be present because for certain kinds of subqueries,
+      the optimizer is not able to see that they were eliminated.
+    */
+    if (!node)
+      continue;
+
+    /* Derived tables are printed inside Explain_table_access objects */
     if (!is_connection_printable_in_json(node->connection_type))
       continue;
 
@@ -756,6 +786,18 @@ bool Explain_node::print_explain_json_cache(Json_writer *writer,
 }
 
 
+bool Explain_node::print_explain_json_subq_materialization(Json_writer *writer,
+                                                           bool is_analyze)
+{
+  if (subq_materialization)
+  {
+    subq_materialization->print_explain_json(writer, is_analyze);
+    return true;
+  }
+  return false;
+}
+
+
 Explain_basic_join::~Explain_basic_join()
 {
   if (join_tabs)
@@ -903,6 +945,8 @@ void Explain_select::print_explain_json(Explain_query *query,
   Json_writer_nesting_guard guard(writer);
   
   bool started_cache= print_explain_json_cache(writer, is_analyze);
+  bool started_subq_mat= print_explain_json_subq_materialization(writer,
+                                                                 is_analyze);
 
   if (message ||
       select_type == pushed_derived_text ||
@@ -1006,6 +1050,8 @@ void Explain_select::print_explain_json(Explain_query *query,
     writer->end_object();
   }
 
+  if (started_subq_mat)
+    writer->end_object();
   if (started_cache)
     writer->end_object();
 }
@@ -1045,11 +1091,11 @@ void Explain_aggr_filesort::print_json_members(Json_writer *writer,
       first= false;
     else
     {
-      str.append(", ");
+      str.append(STRING_WITH_LEN(", "));
     }
     append_item_to_str(&str, item);
     if (*direction == ORDER::ORDER_DESC)
-      str.append(" desc");
+      str.append(STRING_WITH_LEN(" desc"));
   }
 
   writer->add_member("sort_key").add_str(str.c_ptr_safe());
@@ -1127,14 +1173,15 @@ void Explain_table_access::fill_key_str(String *key_str, bool is_json) const
   CHARSET_INFO *cs= system_charset_info;
   bool is_hj= (type == JT_HASH || type == JT_HASH_NEXT || 
                type == JT_HASH_RANGE || type == JT_HASH_INDEX_MERGE);
-  const char *hash_key_prefix= "#hash#";
+  LEX_CSTRING hash_key_prefix= { STRING_WITH_LEN("#hash#") };
+  const char *key_name;
 
-  if (key.get_key_name())
+  if ((key_name= key.get_key_name()))
   {
     if (is_hj)
-      key_str->append(hash_key_prefix, strlen(hash_key_prefix), cs);
+      key_str->append(hash_key_prefix.str, hash_key_prefix.length, cs);
 
-    key_str->append(key.get_key_name());
+    key_str->append(key_name, strlen(key_name));
 
     if (is_hj && type != JT_HASH)
       key_str->append(':');
@@ -1150,7 +1197,10 @@ void Explain_table_access::fill_key_str(String *key_str, bool is_json) const
     key_str->append(buf2);
   }
   if (type == JT_HASH_NEXT)
-    key_str->append(hash_next_key.get_key_name());
+  {
+    key_name= hash_next_key.get_key_name();
+    key_str->append(key_name, strlen(key_name));
+  }
 }
 
 
@@ -1292,8 +1342,8 @@ int Explain_table_access::print_explain(select_result_sink *output, uint8 explai
     push_str(thd, &item_list, join_type_str[type]);
   else
   {
-    join_type_buf.append(join_type_str[type]);
-    join_type_buf.append("|filter");
+    join_type_buf.append(join_type_str[type], strlen(join_type_str[type]));
+    join_type_buf.append(STRING_WITH_LEN("|filter"));
     item_list.push_back(new (mem_root)
                         Item_string_sys(thd, join_type_buf.ptr(),
                                              join_type_buf.length()),
@@ -1313,7 +1363,7 @@ int Explain_table_access::print_explain(select_result_sink *output, uint8 explai
 
   if (rowid_filter)
   {
-    key_str.append("|");
+    key_str.append('|');
     StringBuffer<64> rowid_key_str;
     rowid_filter->quick->print_key(&rowid_key_str);
     key_str.append(rowid_key_str);
@@ -1356,10 +1406,10 @@ int Explain_table_access::print_explain(select_result_sink *output, uint8 explai
 
     if (rowid_filter)
     {
-      rows_str.append(" (");
+      rows_str.append(STRING_WITH_LEN(" ("));
       rows_str.append_ulonglong((ulonglong) (round(rowid_filter->selectivity *
 						   100.0)));
-      rows_str.append("%)");
+      rows_str.append(STRING_WITH_LEN("%)"));
     }
     item_list.push_back(new (mem_root)
                         Item_string_sys(thd, rows_str.ptr(),
@@ -1382,13 +1432,13 @@ int Explain_table_access::print_explain(select_result_sink *output, uint8 explai
       Item_float *fl= new (mem_root) Item_float(thd, avg_rows, 2);
       String tmp;
       String *res= fl->val_str(&tmp);
-      r_rows_str.append(res->ptr());
+      r_rows_str.append(*res);
       if (rowid_filter)
       {
-        r_rows_str.append(" (");
+        r_rows_str.append(STRING_WITH_LEN(" ("));
         r_rows_str.append_ulonglong(
 	  (ulonglong) (rowid_filter->tracker->get_r_selectivity_pct() * 100.0));
-        r_rows_str.append("%)");
+        r_rows_str.append(STRING_WITH_LEN("%)"));
       }
       item_list.push_back(new (mem_root)
                           Item_string_sys(thd, r_rows_str.ptr(),
@@ -1637,6 +1687,9 @@ void Explain_table_access::tag_to_json(Json_writer *writer, enum explain_extra_t
     case ET_DISTINCT:
       writer->add_member("distinct").add_bool(true);
       break;
+    case ET_TABLE_FUNCTION:
+      writer->add_member("table_function").add_str("json_table");
+      break;
 
     default:
       DBUG_ASSERT(0);
@@ -1665,6 +1718,7 @@ void Explain_rowid_filter::print_explain_json(Explain_query *query,
   if (is_analyze)
   {
     writer->add_member("r_rows").add_double(tracker->get_container_elements());
+    writer->add_member("r_lookups").add_ll(tracker->get_container_lookups());
     writer->add_member("r_selectivity_pct").
       add_double(tracker->get_r_selectivity_pct() * 100.0);
     writer->add_member("r_buffer_size").
@@ -1675,6 +1729,28 @@ void Explain_rowid_filter::print_explain_json(Explain_query *query,
   writer->end_object(); // rowid_filter
 }
 
+static void trace_engine_stats(handler *file, Json_writer *writer)
+{
+  if (file && file->handler_stats)
+  {
+    ha_handler_stats *hs= file->handler_stats;
+    writer->add_member("r_engine_stats").start_object();
+    if (hs->pages_accessed)
+      writer->add_member("pages_accessed").add_ull(hs->pages_accessed);
+    if (hs->pages_updated)
+      writer->add_member("pages_updated").add_ull(hs->pages_updated);
+    if (hs->pages_read_count)
+      writer->add_member("pages_read_count").add_ull(hs->pages_read_count);
+    if (hs->pages_read_time)
+      writer->add_member("pages_read_time_ms").
+        add_double(hs->pages_read_time * 1000. / timer_tracker_frequency());
+    if (hs->pages_prefetched)
+      writer->add_member("pages_prefetch_read_count").add_ull(hs->pages_prefetched);
+    if (hs->undo_records_read)
+      writer->add_member("old_rows_read").add_ull(hs->undo_records_read);
+    writer->end_object();
+  }
+}
 
 void Explain_table_access::print_explain_json(Explain_query *query,
                                               Json_writer *writer,
@@ -1816,6 +1892,7 @@ void Explain_table_access::print_explain_json(Explain_query *query,
       writer->add_member("r_table_time_ms").add_double(total_time);
       writer->add_member("r_other_time_ms").add_double(extra_time_tracker.get_time_ms());
     }
+    trace_engine_stats(handler_for_stats, writer);
   }
 
   /* `filtered` */
@@ -1872,12 +1949,40 @@ void Explain_table_access::print_explain_json(Explain_query *query,
 
     if (is_analyze)
     {
-      //writer->add_member("r_loops").add_ll(jbuf_tracker.get_loops());
+      writer->add_member("r_loops").add_ll(jbuf_loops_tracker.get_loops());
+
       writer->add_member("r_filtered");
       if (jbuf_tracker.has_scans())
         writer->add_double(jbuf_tracker.get_filtered_after_where()*100.0);
       else
         writer->add_null();
+
+      writer->add_member("r_unpack_time_ms");
+      writer->add_double(jbuf_unpack_tracker.get_time_ms());
+      DBUG_EXECUTE_IF("analyze_print_r_unpack_ops",
+                      {
+                        writer->add_member("r_unpack_ops");
+                        writer->add_ull(jbuf_unpack_tracker.get_loops());
+                      });
+
+      writer->add_member("r_other_time_ms").
+        add_double(jbuf_extra_time_tracker.get_time_ms());
+      /*
+        effective_rows is average number of matches we got for an incoming
+        row. The row is stored in the join buffer and then is read
+        from there, possibly multiple times. We can't count this number
+        directly. Infer it as:
+         total_number_of_row_combinations_considered / r_loops.
+      */
+      writer->add_member("r_effective_rows");
+      if (jbuf_loops_tracker.has_scans())
+      {
+        double loops= (double)jbuf_loops_tracker.get_loops();
+        double row_combinations= (double)jbuf_tracker.r_rows;
+        writer->add_double(row_combinations / loops);
+      }
+      else
+        writer->add_null();
     }
   }
 
@@ -1928,42 +2033,41 @@ void Explain_table_access::print_explain_json(Explain_query *query,
   sql_explain.h
 */
 
-const char * extra_tag_text[]=
+const LEX_CSTRING extra_tag_text[]=
 {
-  "ET_none",
-  "Using index condition",
-  "Using index condition(BKA)",
-  "Using ", // special handling
-  "Range checked for each record (index map: 0x", // special handling
-  "Using where with pushed condition",
-  "Using where",
-  "Not exists",
+  { STRING_WITH_LEN("ET_none") },
+  { STRING_WITH_LEN("Using index condition") },
+  { STRING_WITH_LEN("Using index condition(BKA)") },
+  { STRING_WITH_LEN("Using ") }, // special handling
+  { STRING_WITH_LEN("Range checked for each record (index map: 0x") },  // special handling
+  { STRING_WITH_LEN("Using where with pushed condition") },
+  { STRING_WITH_LEN("Using where") },
+  { STRING_WITH_LEN("Not exists") },
   
-  "Using index",
-  "Full scan on NULL key",
-  "Skip_open_table",
-  "Open_frm_only",
-  "Open_full_table", 
+  { STRING_WITH_LEN("Using index") },
+  { STRING_WITH_LEN("Full scan on NULL key") },
+  { STRING_WITH_LEN("Skip_open_table") },
+  { STRING_WITH_LEN("Open_frm_only") },
+  { STRING_WITH_LEN("Open_full_table") },
 
-  "Scanned 0 databases",
-  "Scanned 1 database",
-  "Scanned all databases",
+  { STRING_WITH_LEN("Scanned 0 databases") },
+  { STRING_WITH_LEN("Scanned 1 database") },
+  { STRING_WITH_LEN("Scanned all databases") },
 
-  "Using index for group-by", // special handling
+  { STRING_WITH_LEN("Using index for group-by") }, // special handling
+  { STRING_WITH_LEN("USING MRR: DONT PRINT ME") }, // special handling
 
-  "USING MRR: DONT PRINT ME", // special handling
+  { STRING_WITH_LEN("Distinct") },
+  { STRING_WITH_LEN("LooseScan") },
+  { STRING_WITH_LEN("Start temporary") },
+  { STRING_WITH_LEN("End temporary") },
+  { STRING_WITH_LEN("FirstMatch") },               // special handling
 
-  "Distinct",
-  "LooseScan",
-  "Start temporary",
-  "End temporary",
-  "FirstMatch", // special handling
+  { STRING_WITH_LEN("Using join buffer") },        // special handling
 
-  "Using join buffer", // special handling 
-
-  "Const row not found",
-  "Unique row not found",
-  "Impossible ON condition",
+  { STRING_WITH_LEN("Const row not found") },
+  { STRING_WITH_LEN("Unique row not found") },
+  { STRING_WITH_LEN("Impossible ON condition") }
 };
 
 
@@ -1983,7 +2087,8 @@ void Explain_table_access::append_tag_name(String *str, enum explain_extra_tag t
       char buf[MAX_KEY / 4 + 1];
       str->append(STRING_WITH_LEN("Range checked for each "
                                    "record (index map: 0x"));
-      str->append(range_checked_fer->keys_map.print(buf));
+      range_checked_fer->keys_map.print(buf);
+      str->append(buf, strlen(buf));
       str->append(')');
       break;
     }
@@ -1997,12 +2102,16 @@ void Explain_table_access::append_tag_name(String *str, enum explain_extra_tag t
       str->append(extra_tag_text[tag]);
 
       str->append(STRING_WITH_LEN(" ("));
-      const char *buffer_type= bka_type.incremental ? "incremental" : "flat";
+      LEX_CSTRING buffer_type;
+      if (bka_type.incremental)
+        buffer_type= { STRING_WITH_LEN("incremental") };
+      else
+        buffer_type= { STRING_WITH_LEN("flat") };
       str->append(buffer_type);
       str->append(STRING_WITH_LEN(", "));
-      str->append(bka_type.join_alg);
+      str->append(bka_type.join_alg, strlen(bka_type.join_alg));
       str->append(STRING_WITH_LEN(" join"));
-      str->append(STRING_WITH_LEN(")"));
+      str->append(')');
       if (bka_type.mrr_type.length())
       {
         str->append(STRING_WITH_LEN("; "));
@@ -2015,9 +2124,9 @@ void Explain_table_access::append_tag_name(String *str, enum explain_extra_tag t
     {
       if (firstmatch_table_name.length())
       {
-        str->append("FirstMatch(");
+        str->append(STRING_WITH_LEN("FirstMatch("));
         str->append(firstmatch_table_name);
-        str->append(")");
+        str->append(')');
       }
       else
         str->append(extra_tag_text[tag]);
@@ -2027,7 +2136,10 @@ void Explain_table_access::append_tag_name(String *str, enum explain_extra_tag t
     {
       str->append(extra_tag_text[tag]);
       if (loose_scan_is_scanning)
-        str->append(" (scanning)");
+        str->append(STRING_WITH_LEN(" (scanning)"));
+      break;
+    case ET_TABLE_FUNCTION:
+      str->append(STRING_WITH_LEN("Table function: json_table"));
       break;
     }
     default:
@@ -2082,13 +2194,16 @@ void Explain_quick_select::print_json(Json_writer *writer)
 
 void Explain_quick_select::print_extra_recursive(String *str)
 {
+  const char *name;
   if (is_basic())
   {
-    str->append(range.get_key_name());
+    name= range.get_key_name();
+    str->append(name, strlen(name));
   }
   else
   {
-    str->append(get_name_by_type());
+    name= get_name_by_type();
+    str->append(name, strlen(name));
     str->append('(');
     List_iterator_fast<Explain_quick_select> it (children);
     Explain_quick_select* child;
@@ -2137,7 +2252,7 @@ void Explain_quick_select::print_key(String *str)
   {
     if (str->length() > 0)
       str->append(',');
-    str->append(range.get_key_name());
+    str->append(range.get_key_name(), strlen(range.get_key_name()));
   }
   else
   {
@@ -2484,6 +2599,8 @@ void Explain_update::print_explain_json(Explain_query *query,
     }
   }
 
+  trace_engine_stats(handler_for_stats, writer);
+
   if (where_cond)
   {
     writer->add_member("attached_condition");
@@ -2658,3 +2775,41 @@ void Explain_range_checked_fer::print_json(Json_writer *writer,
     writer->end_object();
   }
 }
+
+
+void Explain_subq_materialization::print_explain_json(Json_writer *writer,
+                                                      bool is_analyze)
+{
+  writer->add_member("materialization").start_object();
+  if (is_analyze)
+    tracker.print_json_members(writer);
+}
+
+
+void Subq_materialization_tracker::print_json_members(Json_writer *writer) const
+{
+  writer->add_member("r_strategy").add_str(get_exec_strategy());
+  if (loops_count)
+    writer->add_member("r_loops").add_ull(loops_count);
+
+  if (index_lookups_count)
+    writer->add_member("r_index_lookups").add_ull(index_lookups_count);
+
+  if (partial_matches_count)
+    writer->add_member("r_partial_matches").add_ull(partial_matches_count);
+
+  if (partial_match_buffer_size)
+  {
+    writer->add_member("r_partial_match_buffer_size").
+            add_size(partial_match_buffer_size);
+  }
+
+  if (partial_match_array_sizes.elements())
+  {
+    writer->add_member("r_partial_match_array_sizes").start_array();
+    for(size_t i= 0; i < partial_match_array_sizes.elements(); i++)
+      writer->add_ull(partial_match_array_sizes[i]);
+    writer->end_array();
+  }
+}
+
