@@ -112,11 +112,12 @@ SQL_HANDLER::~SQL_HANDLER()
     Pointer to the TABLE_LIST struct.
 */
 
-static char *mysql_ha_hash_get_key(SQL_HANDLER *table, size_t *key_len,
-                                   my_bool first __attribute__((unused)))
+static const uchar *mysql_ha_hash_get_key(const void *table_, size_t *key_len,
+                                          my_bool)
 {
+  auto table= static_cast<const SQL_HANDLER *>(table_);
   *key_len= table->handler_name.length + 1 ; /* include '\0' in comparisons */
-  return (char*) table->handler_name.str;
+  return reinterpret_cast<const uchar *>(table->handler_name.str);
 }
 
 
@@ -134,9 +135,9 @@ static char *mysql_ha_hash_get_key(SQL_HANDLER *table, size_t *key_len,
     Nothing
 */
 
-static void mysql_ha_hash_free(SQL_HANDLER *table)
+static void mysql_ha_hash_free(void *table)
 {
-  delete table;
+  delete static_cast<SQL_HANDLER *>(table);
 }
 
 static void mysql_ha_close_childs(THD *thd, TABLE_LIST *current_table_list,
@@ -291,8 +292,7 @@ bool mysql_ha_open(THD *thd, TABLE_LIST *tables, SQL_HANDLER *reopen)
     */
     if (my_hash_init(key_memory_THD_handler_tables_hash,
                      &thd->handler_tables_hash, &my_charset_latin1,
-                     HANDLER_TABLES_HASH_SIZE, 0, 0, (my_hash_get_key)
-                     mysql_ha_hash_get_key, (my_hash_free_key)
+                     HANDLER_TABLES_HASH_SIZE, 0, 0, mysql_ha_hash_get_key,
                      mysql_ha_hash_free, 0))
     {
       DBUG_PRINT("exit",("ERROR"));
@@ -616,16 +616,29 @@ static SQL_HANDLER *mysql_ha_find_handler(THD *thd, const LEX_CSTRING *name)
 static bool
 mysql_ha_fix_cond_and_key(SQL_HANDLER *handler, 
                           enum enum_ha_read_modes mode, const char *keyname,
-                          List<Item> *key_expr, enum ha_rkey_function ha_rkey_mode,
+                          List<Item> *key_expr,
+                          enum ha_rkey_function ha_rkey_mode,
                           Item *cond, bool in_prepare)
 {
   THD *thd= handler->thd;
   TABLE *table= handler->table;
   if (cond)
   {
+    bool ret;
+    Item::vcol_func_processor_result res;
+
     /* This can only be true for temp tables */
     if (table->query_id != thd->query_id)
       cond->cleanup();                          // File was reopened
+
+    ret= cond->walk(&Item::check_handler_func_processor, 0, &res);
+    if (ret || res.errors)
+    {
+      my_error(ER_GENERATED_COLUMN_FUNCTION_IS_NOT_ALLOWED, MYF(0), res.name,
+               "WHERE", "HANDLER");
+      return 1;                                 // ROWNUM() used
+    }
+    thd->where= THD_WHERE::WHERE_CLAUSE;
     if (cond->fix_fields_if_needed_for_bool(thd, &cond))
       return 1;
   }
@@ -647,28 +660,26 @@ mysql_ha_fix_cond_and_key(SQL_HANDLER *handler,
       }
     }
 
+    const KEY *c_key= table->s->key_info + handler->keyno;
+    if (c_key->algorithm == HA_KEY_ALG_FULLTEXT ||
+        (ha_rkey_mode != HA_READ_KEY_EXACT &&
+         (table->file->index_flags(handler->keyno, 0, TRUE) &
+          (HA_READ_NEXT | HA_READ_PREV | HA_READ_RANGE)) == 0))
+    {
+      my_error(ER_KEY_DOESNT_SUPPORT, MYF(0),
+               table->file->index_type(handler->keyno), c_key->name.str);
+      return 1;
+    }
+
     /* Check key parts */
     if (mode == RKEY)
     {
-      TABLE *table= handler->table;
       KEY *keyinfo= table->key_info + handler->keyno;
       KEY_PART_INFO *key_part= keyinfo->key_part;
       List_iterator<Item> it_ke(*key_expr);
       Item *item;
       key_part_map keypart_map;
       uint key_len;
-      const KEY *c_key= table->s->key_info + handler->keyno;
-
-      if ((c_key->flags & HA_SPATIAL) ||
-           c_key->algorithm == HA_KEY_ALG_FULLTEXT ||
-          (ha_rkey_mode != HA_READ_KEY_EXACT &&
-           (table->file->index_flags(handler->keyno, 0, TRUE) &
-            (HA_READ_NEXT | HA_READ_PREV | HA_READ_RANGE)) == 0))
-      {
-        my_error(ER_KEY_DOESNT_SUPPORT, MYF(0),
-                 table->file->index_type(handler->keyno), keyinfo->name.str);
-        return 1;
-      }
 
       if (key_expr->elements > keyinfo->user_defined_key_parts)
       {
@@ -686,6 +697,7 @@ mysql_ha_fix_cond_and_key(SQL_HANDLER *handler,
         return 1;
       }
 
+      thd->where= THD_WHERE::HANDLER_STATEMENT;
       for (keypart_map= key_len=0 ; (item=it_ke++) ; key_part++)
       {
 	/* note that 'item' can be changed by fix_fields() call */
@@ -702,7 +714,7 @@ mysql_ha_fix_cond_and_key(SQL_HANDLER *handler,
           MY_BITMAP *old_map= dbug_tmp_use_all_columns(table, &table->write_set);
           int res= item->save_in_field(key_part->field, 1);
           dbug_tmp_restore_column_map(&table->write_set, old_map);
-          if (res)
+          if (res < 0 || thd->is_error())
             return 1;
         }
         key_len+= key_part->store_length;
@@ -977,7 +989,7 @@ retry:
       }
       goto ok;
     }
-    if (cond && !cond->val_int())
+    if (cond && !cond->val_bool())
     {
       if (thd->is_error())
         goto err;
@@ -1023,7 +1035,8 @@ err0:
 SQL_HANDLER *mysql_ha_read_prepare(THD *thd, TABLE_LIST *tables,
                                    enum enum_ha_read_modes mode,
                                    const char *keyname,
-                                   List<Item> *key_expr, enum ha_rkey_function ha_rkey_mode,
+                                   List<Item> *key_expr,
+                                   enum ha_rkey_function ha_rkey_mode,
                                    Item *cond)
 {
   SQL_HANDLER *handler;
