@@ -23,23 +23,35 @@
 typedef List<Item> List_item;
 typedef struct st_copy_info COPY_INFO;
 
-bool mysql_prepare_insert(THD *thd, TABLE_LIST *table_list, TABLE *table,
-                          List<Item> &fields, List_item *values,
-                          List<Item> &update_fields,
-                          List<Item> &update_values, enum_duplicates duplic,
-                          COND **where, bool select_insert);
+int mysql_prepare_insert(THD *thd, TABLE_LIST *table_list,
+                         List<Item> &fields, List_item *values,
+                         List<Item> &update_fields,
+                         List<Item> &update_values, enum_duplicates duplic,
+                         COND **where, bool select_insert, bool * const cache_results);
 bool mysql_insert(THD *thd,TABLE_LIST *table,List<Item> &fields,
                   List<List_item> &values, List<Item> &update_fields,
                   List<Item> &update_values, enum_duplicates flag,
-                  bool ignore);
+                  bool ignore, select_result* result);
 void upgrade_lock_type_for_insert(THD *thd, thr_lock_type *lock_type,
                                   enum_duplicates duplic,
                                   bool is_multi_insert);
 int check_that_all_fields_are_given_values(THD *thd, TABLE *entry,
                                            TABLE_LIST *table_list);
 int vers_insert_history_row(TABLE *table);
-int write_record(THD *thd, TABLE *table, COPY_INFO *info);
+int check_duplic_insert_without_overlaps(THD *thd, TABLE *table,
+                                         enum_duplicates duplic);
+int write_record(THD *thd, TABLE *table, COPY_INFO *info,
+                 select_result *returning= NULL);
 void kill_delayed_threads(void);
+bool binlog_create_table(THD *thd, TABLE *table, bool replace);
+bool binlog_drop_table(THD *thd, TABLE *table);
+
+static inline void restore_default_record_for_insert(TABLE *t)
+{
+  restore_record(t,s->default_values);
+  if (t->triggers)
+    t->triggers->default_extra_null_bitmap();
+}
 
 #ifdef EMBEDDED_LIBRARY
 inline void kill_delayed_threads(void) {}
