@@ -24,19 +24,18 @@
 
 namespace idbdatafile
 {
-
 class UnbufferedFileFactory : public FileFactoryBase
 {
-public:
-    /* virtual */
-    IDBDataFile* open(const char* fname, const char* mode, unsigned opts, unsigned colWidth);
+ public:
+  /* virtual */
+  IDBDataFile* open(const char* fname, const char* mode, unsigned opts, unsigned colWidth);
 };
 
-inline
-IDBDataFile* UnbufferedFileFactory::open(const char* fname, const char* mode, unsigned opts, unsigned /*colWidth*/)
+inline IDBDataFile* UnbufferedFileFactory::open(const char* fname, const char* mode, unsigned opts,
+                                                unsigned /*colWidth*/)
 {
-    return new UnbufferedFile(fname, mode, opts);
+  return new UnbufferedFile(fname, mode, opts);
 }
 
-}
+}  // namespace idbdatafile
 #endif /* UNBUFFEREDFILEFACTORY_H_ */
