@@ -1,7 +1,7 @@
 /*****************************************************************************
 
 Copyright (c) 2013, 2016, Oracle and/or its affiliates. All Rights Reserved.
-Copyright (c) 2017, MariaDB Corporation.
+Copyright (c) 2017, 2021, MariaDB Corporation.
 
 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software
@@ -50,7 +50,6 @@ public:
 	Tablespace()
 		:
 		m_files(),
-		m_name(),
 		m_space_id(ULINT_UNDEFINED),
 		m_path(),
 		m_flags(),
@@ -79,9 +78,6 @@ public:
 	/** Data file iterator */
 	iterator end() { return m_files.end(); }
 
-	void set_name(const char* name) { m_name = name; }
-	const char* name() const { return m_name; }
-
 	/** Set tablespace path and filename members.
 	@param[in]	path	where tablespace file(s) resides
 	@param[in]	len	length of the file path */
@@ -90,8 +86,6 @@ public:
 		ut_ad(m_path == NULL);
 		m_path = mem_strdupl(path, len);
 		ut_ad(m_path != NULL);
-
-		os_normalize_path(m_path);
 	}
 
 	/** Set tablespace path and filename members.
@@ -127,7 +121,7 @@ public:
 	@param[in]	fsp_flags	tablespace flags */
 	void set_flags(ulint fsp_flags)
 	{
-		ut_ad(fsp_flags_is_valid(fsp_flags, false));
+		ut_ad(fil_space_t::is_valid_flags(fsp_flags, false));
 		m_flags = fsp_flags;
 	}
 
@@ -163,9 +157,9 @@ public:
 	void shutdown();
 
 	/** @return the sum of the file sizes of each Datafile */
-	ulint get_sum_of_sizes() const
+	uint32_t get_sum_of_sizes() const
 	{
-		ulint	sum = 0;
+		uint32_t sum = 0;
 
 		for (const_iterator it = begin(); it != end(); ++it) {
 			sum += it->m_size;
@@ -218,9 +212,6 @@ private:
 
 	/* DATA MEMBERS */
 
-	/** Name of the tablespace. */
-	const char*	m_name;
-
 	/** Tablespace ID */
 	ulint		m_space_id;
 
