@@ -163,6 +163,9 @@ struct fts_token_t;
 struct fts_doc_ids_t;
 struct fts_index_cache_t;
 
+/** Compare two DOC_ID. */
+int fts_doc_id_cmp(const void *p1, const void *p2)
+  __attribute__((nonnull, warn_unused_result));
 
 /** Initialize the "fts_table" for internal query into FTS auxiliary
 tables */
@@ -198,9 +201,9 @@ struct fts_trx_t {
 
 /** Information required for transaction savepoint handling. */
 struct fts_savepoint_t {
-	char*		name;		/*!< First entry is always NULL, the
-					default instance. Otherwise the name
-					of the savepoint */
+	const void*	name;		/*!< First entry is always NULL, the
+					default instance. Otherwise the
+					savepoint */
 
 	ib_rbt_t*	tables;		/*!< Modified FTS tables */
 };
@@ -412,6 +415,9 @@ inline void fts_doc_ids_free(fts_doc_ids_t* doc_ids)
 	mem_heap_free(static_cast<mem_heap_t*>(doc_ids->self_heap->arg));
 }
 
+/** Sort an array of doc_id */
+void fts_doc_ids_sort(ib_vector_t *doc_ids);
+
 /******************************************************************//**
 Notify the FTS system about an operation on an FTS-indexed table. */
 void
@@ -610,14 +616,6 @@ fts_create(
 	dict_table_t*	table);			/*!< out: table with FTS
 						indexes */
 
-/**********************************************************************//**
-Free the FTS resources. */
-void
-fts_free(
-/*=====*/
-	dict_table_t*   table);			/*!< in/out: table with
-						FTS indexes */
-
 /*********************************************************************//**
 Run OPTIMIZE on the given table.
 @return DB_SUCCESS if all OK */
@@ -656,13 +654,19 @@ fts_optimize_remove_table(
 void
 fts_optimize_shutdown();
 
+/** Send sync fts cache for the table.
+@param[in]	table	table to sync */
+void
+fts_optimize_request_sync_table(
+	dict_table_t*	table);
+
 /**********************************************************************//**
 Take a FTS savepoint. */
 void
 fts_savepoint_take(
 /*===============*/
 	fts_trx_t*	fts_trx,		/*!< in: fts transaction */
-	const char*	name);			/*!< in: savepoint name */
+	const void*	name);			/*!< in: savepoint */
 
 /**********************************************************************//**
 Refresh last statement savepoint. */
@@ -677,7 +681,7 @@ void
 fts_savepoint_release(
 /*==================*/
 	trx_t*		trx,			/*!< in: transaction */
-	const char*	name);			/*!< in: savepoint name */
+	const void*	name);			/*!< in: savepoint */
 
 /** Clear cache.
 @param[in,out]	cache	fts cache */
@@ -698,7 +702,7 @@ void
 fts_savepoint_rollback(
 /*===================*/
 	trx_t*		trx,			/*!< in: transaction */
-	const char*	name);			/*!< in: savepoint name */
+	const void*	name);			/*!< in: savepoint */
 
 /*********************************************************************//**
 Rollback to and including savepoint indentified by name. */
@@ -710,8 +714,9 @@ fts_savepoint_rollback_last_stmt(
 /** Run SYNC on the table, i.e., write out data from the cache to the
 FTS auxiliary INDEX table and clear the cache at the end.
 @param[in,out]	table		fts table
+@param[in]	wait		whether to wait for existing sync to finish
 @return DB_SUCCESS on success, error code on failure. */
-dberr_t fts_sync_table(dict_table_t* table);
+dberr_t fts_sync_table(dict_table_t* table, bool wait = true);
 
 /****************************************************************//**
 Create an FTS index cache. */
@@ -811,15 +816,14 @@ fts_get_max_doc_id(
 /*===============*/
 	dict_table_t*	table);			/*!< in: user table */
 
-/******************************************************************//**
-Check whether user supplied stopword table exists and is of
-the right format.
-@return the stopword column charset if qualifies */
-CHARSET_INFO*
-fts_valid_stopword_table(
-/*=====================*/
-	const char*	stopword_table_name);	/*!< in: Stopword table
-						name */
+/** Check whether a stopword table is in the right format.
+@param stopword_table_name   table name
+@param row_end   name of the system-versioning end column, or "value"
+@return the stopword column charset
+@retval NULL if the table does not exist or qualify */
+CHARSET_INFO *fts_valid_stopword_table(const char *stopword_table_name,
+                                       const char **row_end= NULL);
+
 /****************************************************************//**
 This function loads specified stopword into FTS cache
 @return true if success */
