@@ -140,7 +140,6 @@ row_undo_node_create(
 	undo->common.type = QUE_NODE_UNDO;
 	undo->common.parent = parent;
 
-	undo->state = UNDO_NODE_FETCH_NEXT;
 	undo->trx = trx;
 
 	btr_pcur_init(&(undo->pcur));
@@ -219,8 +218,7 @@ row_undo_search_clust_to_pcur(
 		log, first mark them DATA_MISSING. So we will know if the
 		value gets updated */
 		if (node->table->n_v_cols
-		    && (node->state == UNDO_UPDATE_PERSISTENT
-			|| node->state == UNDO_UPDATE_TEMPORARY)
+		    && !trx_undo_roll_ptr_is_insert(node->roll_ptr)
 		    && !(node->cmpl_info & UPD_NODE_NO_ORD_CHANGE)) {
 			for (ulint i = 0;
 			     i < dict_table_get_n_v_cols(node->table); i++) {
@@ -256,38 +254,24 @@ func_exit:
 	return(found);
 }
 
-/** Try to truncate the undo logs.
-@param[in,out]	trx	transaction */
-static void row_undo_try_truncate(trx_t* trx)
-{
-	if (trx_undo_t*	undo = trx->rsegs.m_redo.undo) {
-		ut_ad(undo->rseg == trx->rsegs.m_redo.rseg);
-		trx_undo_truncate_end(*undo, trx->undo_no, false);
-	}
-
-	if (trx_undo_t* undo = trx->rsegs.m_noredo.undo) {
-		ut_ad(undo->rseg == trx->rsegs.m_noredo.rseg);
-		trx_undo_truncate_end(*undo, trx->undo_no, true);
-	}
-}
-
 /** Get the latest undo log record for rollback.
 @param[in,out]	node		rollback context
-@return	whether an undo log record was fetched */
-static bool row_undo_rec_get(undo_node_t* node)
+@return	undo block for the undo log record
+@retval nullptr if no undo log record was fetched */
+static buf_block_t* row_undo_rec_get(undo_node_t* node)
 {
 	trx_t* trx = node->trx;
 
 	if (trx->pages_undone) {
 		trx->pages_undone = 0;
-		row_undo_try_truncate(trx);
+		trx_undo_try_truncate(*trx);
 	}
 
 	trx_undo_t*	undo	= NULL;
 	trx_undo_t*	update	= trx->rsegs.m_redo.undo;
 	trx_undo_t*	temp	= trx->rsegs.m_noredo.undo;
 	const undo_no_t	limit	= trx->roll_limit;
-	bool		is_temp = false;
+	node->is_temp = false;
 
 	ut_ad(!update || !temp || update->empty() || temp->empty()
 	      || update->top_undo_no != temp->top_undo_no);
@@ -303,32 +287,38 @@ static bool row_undo_rec_get(undo_node_t* node)
 	if (temp && !temp->empty() && temp->top_undo_no >= limit) {
 		if (!undo || undo->top_undo_no < temp->top_undo_no) {
 			undo = temp;
-			is_temp = true;
+			node->is_temp = true;
 		}
 	}
 
 	if (undo == NULL) {
-		row_undo_try_truncate(trx);
+		trx_undo_try_truncate(*trx);
 		/* Mark any ROLLBACK TO SAVEPOINT completed, so that
 		if the transaction object is committed and reused
 		later, we will default to a full ROLLBACK. */
 		trx->roll_limit = 0;
 		trx->in_rollback = false;
-		return false;
+		return nullptr;
 	}
 
 	ut_ad(!undo->empty());
 	ut_ad(limit <= undo->top_undo_no);
 
 	node->roll_ptr = trx_undo_build_roll_ptr(
-		false, trx_sys.rseg_id(undo->rseg, !is_temp),
+		false, trx_sys.rseg_id(undo->rseg, !node->is_temp),
 		undo->top_page_no, undo->top_offset);
 
 	mtr_t	mtr;
 	mtr.start();
 
-	buf_block_t* undo_page = trx_undo_page_get_s_latched(
-		page_id_t(undo->rseg->space->id, undo->top_page_no), &mtr);
+	buf_block_t* undo_page = buf_page_get(
+		page_id_t(undo->rseg->space->id, undo->top_page_no),
+		0, RW_S_LATCH, &mtr);
+	if (!undo_page) {
+		return nullptr;
+	}
+
+	buf_page_make_young_if_needed(&undo_page->page);
 
 	uint16_t offset = undo->top_offset;
 
@@ -341,7 +331,9 @@ static bool row_undo_rec_get(undo_node_t* node)
 		}
 
 		undo->top_page_no = prev_page->page.id().page_no();
-		undo->top_offset  = page_offset(prev_rec);
+		undo->top_offset = uint16_t(prev_rec - prev_page->page.frame);
+		ut_ad(prev_rec - prev_page->page.frame
+		      == page_offset(prev_rec));
 		undo->top_undo_no = trx_undo_rec_get_undo_no(prev_rec);
 		ut_ad(!undo->empty());
 	} else {
@@ -349,11 +341,20 @@ static bool row_undo_rec_get(undo_node_t* node)
 		ut_ad(undo->empty());
 	}
 
-	node->undo_rec = trx_undo_rec_copy(undo_page->page.frame + offset,
-					   node->heap);
+	undo_page->fix();
 	mtr.commit();
 
-	switch (trx_undo_rec_get_type(node->undo_rec)) {
+	node->undo_rec = undo_page->page.frame + offset;
+
+	const size_t end = mach_read_from_2(node->undo_rec);
+	if (UNIV_UNLIKELY(end <= offset
+			  || end >= srv_page_size - FIL_PAGE_DATA_END)) {
+		undo_page->unfix();
+		node->undo_rec = nullptr;
+		return nullptr;
+	}
+
+	switch (node->undo_rec[2] & (TRX_UNDO_CMPL_INFO_MULT - 1)) {
 	case TRX_UNDO_INSERT_METADATA:
 		/* This record type was introduced in MDEV-11369
 		instant ADD COLUMN, which was implemented after
@@ -367,18 +368,11 @@ static bool row_undo_rec_get(undo_node_t* node)
 	case TRX_UNDO_INSERT_REC:
 	case TRX_UNDO_EMPTY:
 		node->roll_ptr |= 1ULL << ROLL_PTR_INSERT_FLAG_POS;
-		node->state = undo == temp
-			? UNDO_INSERT_TEMPORARY : UNDO_INSERT_PERSISTENT;
-		break;
-	default:
-		node->state = undo == temp
-			? UNDO_UPDATE_TEMPORARY : UNDO_UPDATE_PERSISTENT;
-		break;
 	}
 
 	trx->undo_no = node->undo_no = trx_undo_rec_get_undo_no(
 		node->undo_rec);
-	return true;
+	return undo_page;
 }
 
 /***********************************************************//**
@@ -395,29 +389,17 @@ row_undo(
 {
 	ut_ad(node->trx->in_rollback);
 
-	if (node->state == UNDO_NODE_FETCH_NEXT && !row_undo_rec_get(node)) {
+	buf_block_t* undo_page = row_undo_rec_get(node);
+
+	if (!undo_page) {
 		/* Rollback completed for this query thread */
 		thr->run_node = que_node_get_parent(node);
 		return DB_SUCCESS;
 	}
 
-	dberr_t err;
-
-	switch (node->state) {
-	case UNDO_INSERT_PERSISTENT:
-	case UNDO_INSERT_TEMPORARY:
-		err = row_undo_ins(node, thr);
-		break;
-	case UNDO_UPDATE_PERSISTENT:
-	case UNDO_UPDATE_TEMPORARY:
-		err = row_undo_mod(node, thr);
-		break;
-	default:
-		ut_ad("wrong state" == 0);
-		err = DB_CORRUPTION;
-	}
-
-	node->state = UNDO_NODE_FETCH_NEXT;
+	dberr_t err = trx_undo_roll_ptr_is_insert(node->roll_ptr)
+		? row_undo_ins(node, thr) : row_undo_mod(node, thr);
+	undo_page->unfix();
 	btr_pcur_close(&(node->pcur));
 
 	mem_heap_empty(node->heap);
