@@ -1,4 +1,5 @@
 /* Copyright (C) 2006 MySQL AB & MySQL Finland AB & TCX DataKonsult AB
+   Copyright (c) 2020, MariaDB Corporation.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -138,6 +139,8 @@ ha_checksum _ma_unique_hash(MARIA_UNIQUEDEF *def, const uchar *record)
     {
       uint tmp_length= _ma_calc_blob_length(keyseg->bit_start,pos);
       memcpy((void*) &pos,pos+keyseg->bit_start,sizeof(char*));
+      if (!pos)
+        pos= (const uchar*) ""; /* hash_sort does not support NULL ptr */
       if (!length || length > tmp_length)
 	length=tmp_length;			/* The whole blob */
     }
@@ -145,9 +148,9 @@ ha_checksum _ma_unique_hash(MARIA_UNIQUEDEF *def, const uchar *record)
     if (type == HA_KEYTYPE_TEXT || type == HA_KEYTYPE_VARTEXT1 ||
         type == HA_KEYTYPE_VARTEXT2)
     {
-      keyseg->charset->coll->hash_sort(keyseg->charset,
-                                       (const uchar*) pos, length, &seed1,
-                                       &seed2);
+      my_ci_hash_sort(keyseg->charset,
+                      (const uchar*) pos, length,
+                      &seed1, &seed2);
       crc+= seed1;
     }
     else
@@ -235,12 +238,27 @@ my_bool _ma_unique_comp(MARIA_UNIQUEDEF *def, const uchar *a, const uchar *b,
       }
       memcpy((void*) &pos_a, pos_a+keyseg->bit_start, sizeof(char*));
       memcpy((void*) &pos_b, pos_b+keyseg->bit_start, sizeof(char*));
+      if (pos_a == 0)
+        pos_a= (const uchar *) ""; /* Avoid UBSAN nullptr-with-offset */
+      if (pos_b == 0)
+        pos_b= (const uchar *) ""; /* Avoid UBSAN nullptr-with-offset */
     }
-    if (type == HA_KEYTYPE_TEXT || type == HA_KEYTYPE_VARTEXT1 ||
-        type == HA_KEYTYPE_VARTEXT2)
+    if (type == HA_KEYTYPE_TEXT/* the CHAR data type*/)
     {
-      if (ha_compare_text(keyseg->charset, pos_a, a_length,
-                          pos_b, b_length, 0))
+      if (ha_compare_char_fixed(keyseg->charset,
+                                pos_a, a_length,
+                                pos_b, b_length,
+                                keyseg->length / keyseg->charset->mbmaxlen,
+                                FALSE/*b_is_prefix*/))
+        return 1;
+    }
+    else if (type == HA_KEYTYPE_VARTEXT1 ||
+             type == HA_KEYTYPE_VARTEXT2)
+    {
+      if (ha_compare_char_varying(keyseg->charset,
+                                  pos_a, a_length,
+                                  pos_b, b_length,
+                                  FALSE/*b_is_prefix*/))
         return 1;
     }
     else
