@@ -11,11 +11,12 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02111-1301 USA */
+   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1335 USA */
 
 #include "ma_fulltext.h"
 #include "ma_rt_index.h"
 #include "trnman.h"
+#include <mysqld_error.h>
 
 /**
    Update an old row in a MARIA table
@@ -35,7 +36,7 @@ int maria_update(register MARIA_HA *info, const uchar *oldrec,
   DBUG_ENTER("maria_update");
 
   DBUG_EXECUTE_IF("maria_pretend_crashed_table_on_usage",
-                  maria_print_error(info->s, HA_ERR_CRASHED);
+                  _ma_print_error(info, HA_ERR_CRASHED, 0);
                   DBUG_RETURN(my_errno= HA_ERR_CRASHED););
   if (!(info->update & HA_STATE_AKTIV))
   {
@@ -113,10 +114,19 @@ int maria_update(register MARIA_HA *info, const uchar *oldrec,
       {
         MARIA_KEY new_key, old_key;
 
-        (*keyinfo->make_key)(info,&new_key, i, new_key_buff, newrec,
-                             pos, info->trn->trid);
-        (*keyinfo->make_key)(info,&old_key, i, old_key_buff,
-                             oldrec, pos, info->cur_row.trid);
+        if (!(*keyinfo->make_key)(info,&new_key, i, new_key_buff, newrec,
+                                  pos, info->trn->trid))
+          goto err;
+
+        if (!(*keyinfo->make_key)(info,&old_key, i, old_key_buff,
+                                  oldrec, pos, info->cur_row.trid))
+        {
+          /*
+            Can't make a key from the record from the table already.
+            Sholdn't happen normally. Table is corrupted.
+          */
+          goto err;
+        }
 
         /* The above changed info->lastkey2. Inform maria_rnext_same(). */
         info->update&= ~HA_STATE_RNEXT_SAME;
@@ -204,7 +214,7 @@ err:
     save_errno= HA_ERR_INTERNAL_ERROR;          /* Should never happen */
 
   if (my_errno == HA_ERR_FOUND_DUPP_KEY || my_errno == HA_ERR_OUT_OF_MEM ||
-      my_errno == HA_ERR_RECORD_FILE_FULL)
+      my_errno == HA_ERR_RECORD_FILE_FULL || my_errno == HA_ERR_NULL_IN_SPATIAL)
   {
     info->errkey= (int) i;
     flag=0;
@@ -217,7 +227,7 @@ err:
 	  if ((flag++ && _ma_ft_del(info,i,new_key_buff,newrec,pos)) ||
 	      _ma_ft_add(info,i,old_key_buff,oldrec,pos))
           {
-            _ma_set_fatal_error(share, my_errno);
+            _ma_set_fatal_error(info, my_errno);
 	    break;
           }
 	}
@@ -232,7 +242,7 @@ err:
 	  if ((flag++ && _ma_ck_delete(info, &new_key)) ||
 	      _ma_ck_write(info, &old_key))
           {
-            _ma_set_fatal_error(share, my_errno);
+            _ma_set_fatal_error(info, my_errno);
 	    break;
           }
 	}
@@ -240,7 +250,7 @@ err:
     } while (i-- != 0);
   }
   else
-    _ma_set_fatal_error(share, save_errno);
+    _ma_set_fatal_error(info, save_errno);
 
   info->update= (HA_STATE_CHANGED | HA_STATE_AKTIV | HA_STATE_ROW_CHANGED |
 		 key_changed);
@@ -248,6 +258,6 @@ err:
  err_end:
   _ma_writeinfo(info, WRITEINFO_UPDATE_KEYFILE);
   if (save_errno == HA_ERR_KEY_NOT_FOUND)
-    _ma_set_fatal_error(share, HA_ERR_CRASHED);
+    _ma_set_fatal_error(info, HA_ERR_CRASHED);
   DBUG_RETURN(my_errno=save_errno);
 } /* maria_update */
