@@ -1,5 +1,5 @@
 /* Copyright (C) 2008-2019 Kentoku Shiba
-   Copyright (C) 2019 MariaDB corp
+   Copyright (C) 2019-2022 MariaDB corp
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
@@ -47,36 +47,22 @@ typedef struct st_spider_param_string_parse
   bool locate_param_def(char*& start_param);
 } SPIDER_PARAM_STRING_PARSE;
 
-uchar *spider_tbl_get_key(
-  SPIDER_SHARE *share,
+const uchar *spider_tbl_get_key(
+  const void *share,
   size_t *length,
-  my_bool not_used __attribute__ ((unused))
+  my_bool
 );
 
-#ifdef WITH_PARTITION_STORAGE_ENGINE
-uchar *spider_pt_share_get_key(
-  SPIDER_PARTITION_SHARE *share,
+const uchar *spider_wide_share_get_key(
+  const void *share,
   size_t *length,
-  my_bool not_used __attribute__ ((unused))
+  my_bool
 );
 
-uchar *spider_pt_handler_share_get_key(
-  SPIDER_PARTITION_HANDLER_SHARE *share,
+const uchar *spider_link_get_key(
+  const void *link_for_hash,
   size_t *length,
-  my_bool not_used __attribute__ ((unused))
-);
-#endif
-
-uchar *spider_link_get_key(
-  SPIDER_LINK_FOR_HASH *link_for_hash,
-  size_t *length,
-  my_bool not_used __attribute__ ((unused))
-);
-
-uchar *spider_ha_get_key(
-  ha_spider *spider,
-  size_t *length,
-  my_bool not_used __attribute__ ((unused))
+  my_bool
 );
 
 int spider_get_server(
@@ -141,18 +127,14 @@ int spider_increase_longlong_list(
 int spider_parse_connect_info(
   SPIDER_SHARE *share,
   TABLE_SHARE *table_share,
-#ifdef WITH_PARTITION_STORAGE_ENGINE
   partition_info *part_info,
-#endif
   uint create_table
 );
 
 int spider_set_connect_info_default(
   SPIDER_SHARE *share,
-#ifdef WITH_PARTITION_STORAGE_ENGINE
   partition_element *part_elem,
   partition_element *sub_elem,
-#endif
   TABLE_SHARE *table_share
 );
 
@@ -177,11 +159,12 @@ void spider_print_keys(
 );
 #endif
 
+void spider_create_conn_key_add_one(int* counter, char** target, char* src);
+
 int spider_create_conn_keys(
   SPIDER_SHARE *share
 );
 
-#ifdef SPIDER_HAS_HASH_VALUE_TYPE
 SPIDER_LGTM_TBLHND_SHARE *spider_get_lgtm_tblhnd_share(
   const char *table_name,
   uint table_name_length,
@@ -190,15 +173,6 @@ SPIDER_LGTM_TBLHND_SHARE *spider_get_lgtm_tblhnd_share(
   bool need_to_create,
   int *error_num
 );
-#else
-SPIDER_LGTM_TBLHND_SHARE *spider_get_lgtm_tblhnd_share(
-  const char *table_name,
-  uint table_name_length,
-  bool locked,
-  bool need_to_create,
-  int *error_num
-);
-#endif
 
 void spider_free_lgtm_tblhnd_share_alloc(
   SPIDER_LGTM_TBLHND_SHARE *lgtm_tblhnd_share,
@@ -208,12 +182,8 @@ void spider_free_lgtm_tblhnd_share_alloc(
 SPIDER_SHARE *spider_create_share(
   const char *table_name,
   TABLE_SHARE *table_share,
-#ifdef WITH_PARTITION_STORAGE_ENGINE
   partition_info *part_info,
-#endif
-#ifdef SPIDER_HAS_HASH_VALUE_TYPE
   my_hash_value_type hash_value,
-#endif
   int *error_num
 );
 
@@ -240,39 +210,37 @@ void spider_update_link_status_for_share(
   long link_status
 );
 
-#ifdef WITH_PARTITION_STORAGE_ENGINE
-SPIDER_PARTITION_SHARE *spider_get_pt_share(
+SPIDER_WIDE_SHARE *spider_get_wide_share(
   SPIDER_SHARE *share,
   TABLE_SHARE *table_share,
   int *error_num
 );
 
-int spider_free_pt_share(
-  SPIDER_PARTITION_SHARE *partition_share
+int spider_free_wide_share(
+  SPIDER_WIDE_SHARE *wide_share
 );
 
-void spider_copy_sts_to_pt_share(
-  SPIDER_PARTITION_SHARE *partition_share,
+void spider_copy_sts_to_wide_share(
+  SPIDER_WIDE_SHARE *wide_share,
   SPIDER_SHARE *share
 );
 
 void spider_copy_sts_to_share(
   SPIDER_SHARE *share,
-  SPIDER_PARTITION_SHARE *partition_share
+  SPIDER_WIDE_SHARE *wide_share
 );
 
-void spider_copy_crd_to_pt_share(
-  SPIDER_PARTITION_SHARE *partition_share,
+void spider_copy_crd_to_wide_share(
+  SPIDER_WIDE_SHARE *wide_share,
   SPIDER_SHARE *share,
   int fields
 );
 
 void spider_copy_crd_to_share(
   SPIDER_SHARE *share,
-  SPIDER_PARTITION_SHARE *partition_share,
+  SPIDER_WIDE_SHARE *wide_share,
   int fields
 );
-#endif
 
 int spider_open_all_tables(
   SPIDER_TRX *trx,
@@ -325,7 +293,6 @@ char *spider_create_table_name_string(
   const char *sub_name
 );
 
-#ifdef WITH_PARTITION_STORAGE_ENGINE
 void spider_get_partition_info(
   const char *table_name,
   uint table_name_length,
@@ -334,7 +301,6 @@ void spider_get_partition_info(
   partition_element **part_elem,
   partition_element **sub_elem
 );
-#endif
 
 int spider_get_sts(
   SPIDER_SHARE *share,
@@ -343,9 +309,7 @@ int spider_get_sts(
   ha_spider *spider,
   double sts_interval,
   int sts_mode,
-#ifdef WITH_PARTITION_STORAGE_ENGINE
   int sts_sync,
-#endif
   int sts_sync_level,
   uint flag
 );
@@ -358,9 +322,7 @@ int spider_get_crd(
   TABLE *table,
   double crd_interval,
   int crd_mode,
-#ifdef WITH_PARTITION_STORAGE_ENGINE
   int crd_sync,
-#endif
   int crd_sync_level
 );
 
@@ -382,14 +344,6 @@ bool spider_check_pk_update(
   TABLE *table
 );
 
-#if defined(HS_HAS_SQLCOM) && defined(HAVE_HANDLERSOCKET)
-#ifdef HANDLER_HAS_DIRECT_UPDATE_ROWS
-bool spider_check_hs_pk_update(
-  ha_spider *spider,
-  key_range *key
-);
-#endif
-#endif
 
 void spider_set_tmp_share_pointer(
   SPIDER_SHARE *tmp_share,
@@ -459,7 +413,6 @@ bool spider_check_direct_order_limit(
   ha_spider *spider
 );
 
-#ifdef HANDLER_HAS_DIRECT_AGGREGATE
 bool spider_all_part_in_order(
   ORDER *order,
   TABLE *table
@@ -469,11 +422,10 @@ Field *spider_field_exchange(
   handler *handler,
   Field *field
 );
-#endif
 
 int spider_set_direct_limit_offset(
-                                   ha_spider*		spider
-                                   );
+  ha_spider *spider
+);
 
 bool spider_check_index_merge(
   TABLE *table,
@@ -481,8 +433,8 @@ bool spider_check_index_merge(
 );
 
 int spider_compare_for_sort(
-  SPIDER_SORT *a,
-  SPIDER_SORT *b
+  const void *a,
+  const void *b
 );
 
 ulong spider_calc_for_sort(
@@ -520,19 +472,12 @@ void spider_free_spider_object_for_share(
   ha_spider **spider
 );
 
-int spider_create_sts_threads(
-  SPIDER_THREAD *spider_thread
+int spider_create_sts_crd_threads(
+  SPIDER_THREAD *spider_thread,
+  bool is_sts
 );
 
-void spider_free_sts_threads(
-  SPIDER_THREAD *spider_thread
-);
-
-int spider_create_crd_threads(
-  SPIDER_THREAD *spider_thread
-);
-
-void spider_free_crd_threads(
+void spider_free_sts_crd_threads(
   SPIDER_THREAD *spider_thread
 );
 
@@ -560,3 +505,9 @@ void spider_table_remove_share_from_crd_thread(
   SPIDER_SHARE *share
 );
 #endif
+uchar *spider_duplicate_char(
+  uchar *dst,
+  uchar esc,
+  uchar *src,
+  uint src_lgt
+);
