@@ -22,16 +22,14 @@
 #define WSREP_TRANSACTION_HPP
 
 #include "provider.hpp"
-#include "server_state.hpp"
 #include "transaction_id.hpp"
 #include "streaming_context.hpp"
 #include "lock.hpp"
 #include "sr_key_set.hpp"
 #include "buffer.hpp"
-#include "client_service.hpp"
 #include "xid.hpp"
 
-#include <cassert>
+#include <iosfwd>
 #include <vector>
 
 namespace wsrep
@@ -40,6 +38,7 @@ namespace wsrep
     class client_state;
     class key;
     class const_buffer;
+    class server_service;
 
     class transaction
     {
@@ -93,7 +92,7 @@ namespace wsrep
         { return (ws_meta_.seqno().is_undefined() == false); }
 
         /**
-         * Return true if any fragments have been succesfully certified
+         * Return true if any fragments have been successfully certified
          * for the transaction.
          */
         bool is_streaming() const
@@ -104,7 +103,7 @@ namespace wsrep
         /**
          * Return number of fragments certified for current statement.
          *
-         * This counts fragments which have been succesfully certified
+         * This counts fragments which have been successfully certified
          * since the construction of object or last after_statement()
          * call.
          *
@@ -143,9 +142,14 @@ namespace wsrep
 
         int xa_replay(wsrep::unique_lock<wsrep::mutex>&);
 
-        bool pa_unsafe() const { return pa_unsafe_; }
-        void pa_unsafe(bool pa_unsafe) { pa_unsafe_ = pa_unsafe; }
-
+        bool pa_unsafe() const { return (flags() & wsrep::provider::flag::pa_unsafe); }
+        void pa_unsafe(bool pa_unsafe) {
+          if (pa_unsafe) {
+            flags(flags() | wsrep::provider::flag::pa_unsafe);
+          } else {
+            flags(flags() & ~wsrep::provider::flag::pa_unsafe);
+          }
+        }
         bool implicit_deps() const { return implicit_deps_; }
         void implicit_deps(bool implicit) { implicit_deps_ = implicit; }
 
@@ -171,11 +175,12 @@ namespace wsrep
 
         int after_row();
 
-        int before_prepare(wsrep::unique_lock<wsrep::mutex>&);
+        int before_prepare(wsrep::unique_lock<wsrep::mutex>&,
+                           const wsrep::provider::seq_cb_t*);
 
         int after_prepare(wsrep::unique_lock<wsrep::mutex>&);
 
-        int before_commit();
+        int before_commit(const wsrep::provider::seq_cb_t*);
 
         int ordered_commit();
 
@@ -188,13 +193,18 @@ namespace wsrep
         int before_statement();
 
         int after_statement();
+        int after_statement(wsrep::unique_lock<wsrep::mutex>&);
+
+        void after_command_must_abort(wsrep::unique_lock<wsrep::mutex>&);
 
         void after_applying();
 
         bool bf_abort(wsrep::unique_lock<wsrep::mutex>& lock,
-                      wsrep::seqno bf_seqno);
+                      wsrep::seqno bf_seqno,
+                      wsrep::client_service&);
         bool total_order_bf_abort(wsrep::unique_lock<wsrep::mutex>&,
-                                  wsrep::seqno bf_seqno);
+                                  wsrep::seqno bf_seqno,
+                                  wsrep::client_service&);
 
         void clone_for_replay(const wsrep::transaction& other);
 
@@ -241,12 +251,16 @@ namespace wsrep
         bool abort_or_interrupt(wsrep::unique_lock<wsrep::mutex>&);
         int streaming_step(wsrep::unique_lock<wsrep::mutex>&, bool force = false);
         int certify_fragment(wsrep::unique_lock<wsrep::mutex>&);
-        int certify_commit(wsrep::unique_lock<wsrep::mutex>&);
+        int certify_commit(wsrep::unique_lock<wsrep::mutex>&,
+                           const wsrep::provider::seq_cb_t*);
         int append_sr_keys_for_commit();
         int release_commit_order(wsrep::unique_lock<wsrep::mutex>&);
+        void remove_fragments_in_storage_service_scope(
+            wsrep::unique_lock<wsrep::mutex>&);
         void streaming_rollback(wsrep::unique_lock<wsrep::mutex>&);
         int replay(wsrep::unique_lock<wsrep::mutex>&);
-        void clear_fragments();
+        void xa_replay_common(wsrep::unique_lock<wsrep::mutex>&);
+        int xa_replay_commit(wsrep::unique_lock<wsrep::mutex>&);
         void cleanup();
         void debug_log_state(const char*) const;
         void debug_log_key_append(const wsrep::key& key) const;
@@ -265,7 +279,6 @@ namespace wsrep
         wsrep::ws_handle ws_handle_;
         wsrep::ws_meta ws_meta_;
         int flags_;
-        bool pa_unsafe_;
         bool implicit_deps_;
         bool certified_;
         size_t fragments_certified_for_statement_;
@@ -273,6 +286,15 @@ namespace wsrep
         wsrep::sr_key_set sr_keys_;
         wsrep::mutable_buffer apply_error_buf_;
         wsrep::xid xid_;
+        bool streaming_rollback_in_progress_;
+        /* This flag tells that the transaction has become immutable
+           against BF aborts. Ideally this would be deduced from transaction
+           state, i.e. s_ordered_commit or s_committed, but the current
+           version of wsrep-lib changes the transaction state to
+           s_ordered_commit too late. Fixing this appears to require
+           too many changes to application using the lib, so boolean flag
+           must do. */
+        bool is_bf_immutable_;
     };
 
     static inline const char* to_c_string(enum wsrep::transaction::state state)
@@ -300,6 +322,8 @@ namespace wsrep
         return to_c_string(state);
     }
 
+    std::ostream& operator<<(std::ostream& os, enum wsrep::transaction::state);
+
 }
 
 #endif // WSREP_TRANSACTION_HPP
